{% extends "!base.html" %}
{% block extrahead %}
  <link rel="preload" href="_static/fonts/JetBrainsMono/JetBrainsMono-Regular.woff2" as="font" type="font/woff2" crossorigin />
  <link rel="preload" href="_static/fonts/JetBrainsMono/JetBrainsMono-Italic.woff2" as="font" type="font/woff2" crossorigin />
  <link rel="preload" href="_static/fonts/JetBrainsMono/JetBrainsMono-Bold.woff2" as="font" type="font/woff2" crossorigin />
  <link rel="preload" href="_static/fonts/JetBrainsMono/JetBrainsMono-BoldItalic.woff2" as="font" type="font/woff2" crossorigin />
  <link rel="preload" href="_static/fonts/Signika/Signika-Regular.woff2" as="font" type="font/woff2" crossorigin />
  <link rel="preload" href="_static/fonts/Signika/Signika-Bold.woff2" as="font" type="font/woff2" crossorigin />
  {{ super() }}
{% endblock %}
