/*
 * The SIL Open Font Licence version 1.1 does not permit using the reserved
 * name for modified versions of a font. Deleting components (to reduce the
 * size of the package I want to bundle) makes the result a modified version.
 * Hence, I have used different names.
 */

@font-face {
  font-family: "JetBrainsMono pysorteddict";
  src: url(../fonts/JetBrainsMono/JetBrainsMono-Regular.woff2);
}

@font-face {
  font-family: "JetBrainsMono pysorteddict";
  font-style: italic;
  src: url(../fonts/JetBrainsMono/JetBrainsMono-Italic.woff2);
}

@font-face {
  font-family: "JetBrainsMono pysorteddict";
  font-weight: bold;
  src: url(../fonts/JetBrainsMono/JetBrainsMono-Bold.woff2);
}

@font-face {
  font-family: "JetBrainsMono pysorteddict";
  font-style: italic;
  font-weight: bold;
  src: url(../fonts/JetBrainsMono/JetBrainsMono-BoldItalic.woff2);
}

@font-face {
  font-family: "Signika pysorteddict";
  src: url(../fonts/Signika/Signika-Regular.woff2);
}

@font-face {
  font-family: "Signika pysorteddict";
  font-weight: bold;
  src: url(../fonts/Signika/Signika-Bold.woff2);
}

::selection {
  background: rgba(39, 180, 166, 0.4);
}

summary {
  color: var(--color-link);
  cursor: pointer;
}

summary:hover {
  text-decoration: underline;
}

.notice {
  background: rgba(39, 180, 166, 0.2);
  border-left: 4px solid rgb(39, 180, 166);
  border-radius: 6px;
  margin: 20px 0px 20px 0px;
  padding: 10px;
}

.warning {
  background: rgba(175, 175, 95, 0.2);
  border-left: 4px solid rgb(175, 175, 95);
  border-radius: 6px;
  margin: 20px 0px 20px 0px;
  padding: 10px;
}

.card-container {
  column-gap: 13%;
  display: flex;
  flex-wrap: wrap;
  justify-content: center;
  row-gap: 32px;
}

.card {
  display: inline-flex;
  flex-direction: column;
  text-align: center;
  transition: background 0.4s ease;
}

.card:hover {
  background: rgba(39, 180, 166, 0.2);
}

.card:active {
  background: rgba(39, 180, 166, 0.6);
}

/*
 * Markdown does not support row and column spans, so I used reStructuredText
 * to draw a table on the performance page. However, the latter does not offer
 * a way to centre the contents of all table cells. This hack does it.
 */
#overview table td {
  text-align: center;
}
